// ConfigView.cpp - Implementation for the ConfigView class
// Provides controls for our screensaver


#include "ConfigView.h"


// Constructor
ConfigView::ConfigView(BRect frame, const char *name,
 int32* new_num_sparks, int32* new_trail_length, float* new_gravity, bool* new_blending)
 : BView(frame, name, 0, B_FOLLOW_ALL),
 num_sparks(new_num_sparks), trail_length(new_trail_length), gravity(new_gravity), blending(new_blending) 
{
  // Nothing to do here...
}
    
    
// AttachedToWindow() gets called when the view is attached to a parent
//  window, which is a good time to set up our menu widgets
void ConfigView::AttachedToWindow()
{
  // Make our background color match the one used by the screensaver
  //  configuration panel
  SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));

  BFont font;  
  font_height f_height;

  float vert_spacing, y_pos;
  float s_width, s_height;


  // Create the configuration menu
  
  // First get some info about the font we're using
  this->GetFont(&font);
  this->GetFontHeight(&f_height);
  vert_spacing = f_height.ascent + f_height.descent + f_height.leading;
  y_pos = 0;

  // Title  
  // Set starting row position
  title_view=new BStringView(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing), B_EMPTY_STRING, CREDITS_TITLE);
  title_view->SetAlignment(B_ALIGN_CENTER);  
  this->AddChild(title_view);
  y_pos += vert_spacing;
    
  // Date / Version
  version_view=new BStringView(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), B_EMPTY_STRING, CREDITS_VERSION);
  version_view->SetAlignment(B_ALIGN_CENTER);  
  this->AddChild(version_view);
  y_pos += vert_spacing;
 
  // Credits
  credits_view=new BStringView(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), B_EMPTY_STRING, CREDITS_AUTHOR);
  credits_view->SetAlignment(B_ALIGN_CENTER);  
  this->AddChild(credits_view);
  y_pos += vert_spacing;
  
  // Spacing between credits and sliders
  y_pos += vert_spacing + vert_spacing;
  
  // Spacing between sliders is a little less...
  vert_spacing = f_height.descent + f_height.leading;
  
  
  // Slider for number of sparks
  num_sparks_slider = new BSlider(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), "const:slider1", "Number of Particles", new BMessage('cvs0'), MIN_SPARKS, MAX_SPARKS);
  num_sparks_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
  num_sparks_slider->SetHashMarkCount(10);
  num_sparks_slider->SetLimitLabels("Fewer", "More");
  num_sparks_slider->SetPosition((float) *num_sparks/MAX_SPARKS); 
  this->AddChild(num_sparks_slider);
  num_sparks_slider->SetTarget(this);
  num_sparks_slider->GetPreferredSize(&s_width, &s_height);
  y_pos += s_height + vert_spacing;



  // Slider for trail length
  trail_length_slider = new BSlider(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), "const:slider2", "Trail Length", new BMessage('cvs1'), MIN_TRAIL_LENGTH, MAX_TRAIL_LENGTH);
  trail_length_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
  trail_length_slider->SetHashMarkCount(10);
  trail_length_slider->SetLimitLabels("Short", "Long");
  trail_length_slider->SetPosition((float) *trail_length/MAX_TRAIL_LENGTH); 
  this->AddChild(trail_length_slider);
  trail_length_slider->SetTarget(this);
  trail_length_slider->GetPreferredSize(&s_width, &s_height);
  y_pos += s_height + vert_spacing;
  
  
  
  // Slider for gravity
  // We make this one go from 1 to 100 to get some granularity -- gravity isn't an int...
  gravity_slider = new BSlider(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), "const:slider3", "Interaction", new BMessage('cvs2'), 1, 200);
  gravity_slider->SetHashMarks(B_HASH_MARKS_BOTTOM);
  gravity_slider->SetHashMarkCount(10);
  gravity_slider->SetLimitLabels("Repulse", "Attract");
  // Some funkiness is needed here to get it to a scale of 0...1 instead of -...+
  // We actually divide the scale in half, the negative side going to repulsion and
  //  the positive side going to attraction
  if(*gravity < 0)
  {
    // Remember to subtract it from 0.5, to flip the direction of increase...
    gravity_slider->SetPosition( 0.5 - ((fabs(*gravity) / fabs(MIN_GRAVITY)) / 2.0) );
  }
  else
  {
    gravity_slider->SetPosition( ((fabs(*gravity) / fabs(MAX_GRAVITY)) / 2.0) + 0.5);
  }
  this->AddChild(gravity_slider);
  gravity_slider->SetTarget(this);
  y_pos += s_height + vert_spacing;


  // Checkbox to enable/disable blending
  blending_checkbox = new BCheckBox(BRect(0, y_pos, this->Bounds().Width(),
   vert_spacing + y_pos), "const:checkbox1", "Enable blending",
   new BMessage('cvb1'), B_FOLLOW_LEFT | B_FOLLOW_TOP, B_WILL_DRAW | B_NAVIGABLE);
  if(*blending)
  {
    blending_checkbox->SetValue(B_CONTROL_ON);
  }
  else
  {
    blending_checkbox->SetValue(B_CONTROL_OFF);
  }
  this->AddChild(blending_checkbox);
  blending_checkbox->SetTarget(this);

}


// Handle messages to sliders, and update our configuration variables with
//  the current values of the sliders
void ConfigView::MessageReceived(BMessage *msg)
{
  switch(msg->what)
  {
    case 'cvs0':
      *num_sparks = (int) (num_sparks_slider->Position() * MAX_SPARKS);
      // Needed because of a slight rounding error
      if(*num_sparks < MIN_SPARKS)
      {
        *num_sparks = MIN_SPARKS;
      }
      if(*num_sparks > MAX_SPARKS)
      {
        *num_sparks = MAX_SPARKS;
      }
      break;

    case 'cvs1':
      *trail_length = (int) (trail_length_slider->Position() * MAX_TRAIL_LENGTH);
      // Needed because of a slight rounding error
      if(*trail_length < MIN_TRAIL_LENGTH)
      {
        *trail_length = MIN_TRAIL_LENGTH;
      }
      if(*trail_length > MAX_TRAIL_LENGTH)
      {
        *trail_length = MAX_TRAIL_LENGTH;
      }
      break;

    case 'cvs2':
      // We actually divide this in half, the negative side devoted to repulsion and the
      //  positive side to attraction
      *gravity = (gravity_slider->Position() - 0.5);
      
      // Needed because of a slight rounding error when slider position = 1.0
      if(*gravity > 0.49)
      {
        *gravity = 0.49;
      }
      
      if(*gravity < 0)
      {
        *gravity = (*gravity * 2.0) * fabs(MIN_GRAVITY);
      }
      else
      {
        *gravity = (*gravity * 2.0) * fabs(MAX_GRAVITY);
      }
      break;

    case 'cvb1':
      if(blending_checkbox->Value() == B_CONTROL_ON)
      {
        *blending = true;
      }
      else
      {
        *blending = false;
      }      
      break;
      
    default:
      break;
  }

  // Respond that we have handled the message that came in (whatever it was)
  //  so that it'll be removed from the message queue
  BView::MessageReceived(msg);
}


