// Sparticles.cpp -- Implementation of the Sparticles class
// Jon Pimble - fatrat@fat-rat.com


#include "Sparticles.h"


extern "C" _EXPORT BScreenSaver *instantiate_screen_saver( BMessage *msg, image_id id )
{
  return new Sparticles( msg, id ); 
} 


// Constructor
// We want to keep the constructor as lightweight as possible, since EVERY screensaver available has its
//  constructor called by the screensaver preferences tool; if any one spends a lot of time initializing
//  inside its constructor, the user has to wait -- all of the time spent in constructors adds up, even
//  if each one is only doing a little bit of work...
Sparticles::Sparticles( BMessage *archive, image_id image ) 
          : BScreenSaver( archive, image ) 
{ 
  // Get initial values from screensaver caller, and make sure they're
  //  within an acceptable range; if not, use default values
  if(archive->FindInt32("num_sparks", &num_sparks) != B_OK)
  {
    num_sparks = DEFAULT_SPARKS;
  }
  if((num_sparks < MIN_SPARKS) || (num_sparks > MAX_SPARKS))
  {
    num_sparks = DEFAULT_SPARKS;
  }

  if(archive->FindInt32("trail_length", &trail_length) != B_OK)
  {
    trail_length = DEFAULT_TRAIL_LENGTH;
  }
  if((trail_length < MIN_TRAIL_LENGTH) || (trail_length > MAX_TRAIL_LENGTH))
  {
    trail_length = DEFAULT_TRAIL_LENGTH;
  }

  if(archive->FindFloat("gravity", &gravity) != B_OK)
  {
    gravity = DEFAULT_GRAVITY;
  }
  if((gravity < MIN_GRAVITY) || (gravity > MAX_GRAVITY))
  {
    gravity = DEFAULT_GRAVITY;
  }
  
  if(archive->FindBool("blending", &blending) != B_OK)
  {
    blending = false;
  }
  
}



// Destructor
Sparticles::~Sparticles()
{
  // Nothing to do...
}


       
// Saves the state of the screensaver (stores config varliables)
status_t Sparticles::SaveState(BMessage *into) const
{
  // Save state of configuration settings
  into->AddInt32("num_sparks", num_sparks);
  into->AddInt32("trail_length", trail_length);
  into->AddFloat("gravity", gravity);
  into->AddBool("blending", blending);
  
  return B_OK;
}



// Setup the menu shown in the screensaver preferences window
// Note that we don't have to bother destroying this later, since the parent view will take care
//  of that for us when it is destroyed
void Sparticles::StartConfig(BView *view)
{       
  configs = new ConfigView(view->Bounds(), "configs", &num_sparks, &trail_length, &gravity, &blending);
  view->AddChild(configs);
}



// Setup screensaver parameters before inital drawing
status_t Sparticles::StartSaver( BView *view, bool preview ) 
{
  // Seed for random number generator
  srand(time(NULL));
  
  // Time between calls to Draw(), in microseconds
  SetTickSize(TICKS);

  // Assume we're *NOT* in DirectDraw unless told otherwise
  dd_mode = false;
    
  // Must return B_OK for screensaver to start
  return B_OK;
}



// Called when/if DirectDraw capability is present
void Sparticles::DirectConnected(direct_buffer_info *info)
{
  // Extract the info we care about from the direct_buffer_info
  pixel_data = (char *) info->bits;
  stride = info->bytes_per_row;
  bpp = info->bits_per_pixel;
  colorspace = info->pixel_format;
  dd_bounds = info->window_bounds;
  
  // Note that we are in direct draw mode only if it's in a colorspace we can handle; otherwise, we'll
  //  just fall back on the standard BeOS drawing rules
  if(
       (colorspace == B_RGB32) || (colorspace == B_RGBA32)
    || (colorspace == B_RGB24)
    || (colorspace == B_RGB16)
    || (colorspace == B_RGB15) || (colorspace == B_RGBA15)
  )
  {
    dd_mode = true;
  }
  else
  {
    dd_mode = false;
  }
}



// Main drawing and animation function
void Sparticles::Draw(BView *view, int32 frame)
{
  // Lots of scratch variables
  uint8 c1, c2, tmp_red, tmp_green, tmp_blue;
  char *tmp_buf = NULL;
  int i, j, x, tmp_pos, tmp_loop_pos, trail_pos, tmp_buf_len;
  float tmp_force, tmp_dist, too_far;
  float width, height, s_width, s_height;  
  int q_width, q_height;
  Vect2D tmp_vect;  
  BRect scratch_rect;


  // Some pre-calculated stuff - figuring these out here saves us from doing a number
  //  of multiplies inside loops
  // Size of screen in real dimensions
  width = view->Bounds().Width();
  height = view->Bounds().Height();
  
  // Size of squares to draw, based on real screen dimensions
  q_width = (int) ((width/1024)* 5);
  q_height = (int) ((height/768) * 5);

  // Dimensions of screen scaled to our coordinates
  s_width = width/MAX_X_POS;
  s_height = height/MAX_Y_POS;
  
  
  // Minimum distace between points to warrant calculating interacting forces
  too_far = fabs(gravity) * MAX_MASS / MIN_PULL;
    

  // Blending and DirectDraw don't mix for now...
  if(blending)
  {
    dd_mode = false;
  }

  
  // Build our color lookup tables during the first frame
  if(frame == 0)
  {
    // Build some lookup tables; these allow us to quickly go from 24bpp (RGB) to 16 or 15bpp
    for(i=0; i < 256; i++)
    {
      // Lookup table for colors for 16-bpp
      red_16_LUT[i] = ( (uint8) ((i / 255.0) * 31) << 3);
      green_h_16_LUT[i] = ( (uint8) ((i / 255.0) * 63) >> 3);
      green_l_16_LUT[i] = (( (uint8) ((i / 255.0) * 63) & 0x07) << 5);
      blue_16_LUT[i] = ( (uint8) ((i / 255.0) * 31));

      // Color lookup table for 15-bpp
      red_15_LUT[i] = ( (uint8) ((i / 255.0) * 31) << 2);
      green_h_15_LUT[i] = ( (uint8) ((i / 255.0) * 31) >> 3);
      green_l_15_LUT[i] = (( (uint8) ((i / 255.0) * 31) & 0x07) << 5);
      // blue for 15bpp is the same as 16bpp, so we don't bother recalculating
    }
  }
  

  // In the first frame, blank the background and initialize the sparks
  // Also check to see if any configs have changed - if they have, re-init
  if( (frame == 0) || (old_num_sparks != num_sparks)
   || (old_trail_length != trail_length) || (old_gravity != gravity) || (old_blending != blending))
  {  
    // Keep track of changed config variables
    old_num_sparks = num_sparks;
    old_trail_length = trail_length;
    old_gravity = gravity;
    old_blending = blending;
    
    // Black out background
    view->SetLowColor(0, 0, 0);
    view->FillRect(view->Bounds(), B_SOLID_LOW);
    
    // Initialize all sparks
    for(i = 0; i < num_sparks; i++)
    {
      init_spark(i, width, height);
    }
  }


  // Calculate spark interactions
  for(i=0; i < (num_sparks-1); i++)
  {
    // Loop for each item in the list
    for(j=i+1; j < num_sparks; j++)
    {
      // Vector representing the distance/direction between two sparks
      tmp_vect.x = sparks[j].x - sparks[i].x;
      tmp_vect.y = sparks[j].y - sparks[i].y;      

      // Calculate the distance (squared) between the sparks
      // Note: we could set our tmp_vect here and use magnitude(), but that would incur an extra
      //  sqrt() call -- we need the distance squared here, not the actual distance
      //  Note: distance = sqrt( (dx*dx) + (dy*dy) )
      tmp_dist = (tmp_vect.x * tmp_vect.x) + (tmp_vect.y * tmp_vect.y);
      
      // Skip all calculations if they are so far apart, no effect will be felt
      //  (skipping this calculation when possible saves a LOT of CPU time)
      // We also skip if the distance is 0.0, since we define the force between the two sparks to be 0
      //  if they're right on top of one another -- this gives them a chance to move away from eachother;
      //  otherwise they tend to sit right on top of the same spot, which looks boring
      if( (tmp_dist > too_far) || (tmp_dist == 0.0) )
      {
        // Do nothing...
      }
      else
      {
        // Unit vector for direction between this spark and the other one
        tmp_vect = tmp_vect.unit();


        // Calculate the force pulling this spark towards the other one
        tmp_force = (gravity * ((float) sparks[j].timeout/MAX_TIMEOUT)*MAX_MASS) / tmp_dist;

        // Add in the direction * force to the velocity of this spark
        sparks[i].velocity += tmp_vect * tmp_force;


        // Calculate the force pulling the other spark towards this one
        // We do this here, because we've already calculated the unit vector, etc; more efficient this way
        tmp_force = (gravity * ((float) sparks[i].timeout/MAX_TIMEOUT)*MAX_MASS) / tmp_dist;

        // Add in the direction * force to the velocity of the other spark
        // Note: we flip the direction of tmp_vect, since this is a pull in the OPPOSITE direction
        sparks[j].velocity += (-tmp_vect) * tmp_force;
      }
    }
  }


  // Now move all of the sparks based on their movement vectors...
  for(i=0; i < num_sparks; i++)
  {
    // Check if spark has strayed out of bounds -- if it has, flip its direction, etc.
    // Note: Extra 10% pull-in is to let more sparks be on-screen; otherwise, more than
    //  half of them tend to stay off-screen
    if( (sparks[i].x < MIN_X_POS) || (sparks[i].y < MIN_Y_POS)
     || (sparks[i].x > MAX_X_POS) || (sparks[i].y > MAX_Y_POS)
    )
    {
      // Pull the spark back towards the center of the area
      tmp_vect.x = (((MAX_X_POS - MIN_X_POS)/2.0) - sparks[i].x);
      tmp_vect.y = (((MAX_Y_POS - MIN_Y_POS)/2.0) - sparks[i].y);
      
      // Pull is 1/10 of minimum velocity (we want this to be small, otherwise the sparks "bounce" off of
      //  the edges of the screen, which looks silly)
      sparks[i].velocity += tmp_vect.unit() * (MIN_VELOCITY/10.0);
    }
    
    
    // Keep sparks from moving too fast or slow
    // Note: having a minimum velocity keeps everything moving and interesting; having a maximum
    //  velocity lets us draw only squares instead of lines to represent the sparks (i.e., we know that
    //  each step in movement along the trail won't be too far from the last)
    if(sparks[i].velocity.magnitude() < MIN_VELOCITY)
    {
      sparks[i].velocity = sparks[i].velocity.unit() * MIN_VELOCITY;
    }
    else
    {
      if(sparks[i].velocity.magnitude() > MAX_VELOCITY)
      {
        sparks[i].velocity = sparks[i].velocity.unit() * MAX_VELOCITY;
      }
    }


    // Move the spark, by adjusting the position based on the velocity vector
    sparks[i].x += sparks[i].velocity.x;
    sparks[i].y += sparks[i].velocity.y;


    // Keep track of spark positions for drawing trails...
    sparks[i].trail_counter++;
    if(sparks[i].trail_counter >= MAX_TRAIL_LENGTH)
    {
      sparks[i].trail_counter = 0;
    }

    // To make up some time, we do the actual on-screen coordinate calculation here
    //  (pulls a multiply-per-square out of the drawing loop)
    
    sparks[i].trail_x[sparks[i].trail_counter] = (int) (sparks[i].x * s_width);
    sparks[i].trail_y[sparks[i].trail_counter] = (int) (sparks[i].y * s_height);

    // *** DD mode optimization
    if(dd_mode)
    {
      // Calculate offset based on colorspace
      switch(colorspace)
      {
        // 32bpp mode
        case B_RGB32:
        case B_RGBA32:
          sparks[i].trail_ofs[sparks[i].trail_counter] =
           (int) pixel_data + ((int)sparks[i].trail_x[sparks[i].trail_counter]*4) + (stride*(int)sparks[i].trail_y[sparks[i].trail_counter]);
          break;

        // 24bpp mode
        case B_RGB24:
          sparks[i].trail_ofs[sparks[i].trail_counter] =
           (int) pixel_data + ((int)sparks[i].trail_x[sparks[i].trail_counter]*3) + (stride*(int)sparks[i].trail_y[sparks[i].trail_counter]);
          break;

        // 16bpp, 15bpp modes
        case B_RGB16:
        case B_RGB15:
        case B_RGBA15:
          sparks[i].trail_ofs[sparks[i].trail_counter] =
           (int) pixel_data + ((int)sparks[i].trail_x[sparks[i].trail_counter]*2) + (stride*(int)sparks[i].trail_y[sparks[i].trail_counter]);
          break;

        // If we make it here... bad things happened...
        default:
          break;
      }
    }
    
  }



  // Draw sparks

  // Wait for the vertical retrace, to make drawing less prone to flicker
  BScreen screen;
  screen.WaitForRetrace();


  // If blending is enabled...
  if(blending)
  {
    view->SetDrawingMode(B_OP_BLEND);
  }

  // If we're in DirectDraw mode, lock the screen, just to be safe
  if(dd_mode)
  {
    tmp_buf = new char[q_width*4];
    memset(tmp_buf, 0, q_width*4);
    locker.Lock();
  }

  // Draw all of the sparks in turn
  for(i=0; i < num_sparks; i++)
  {
    // Start at the end of the trail
    trail_pos = sparks[i].trail_counter - sparks[i].trail_length;
    if(trail_pos < 0)
    {
      trail_pos += MAX_TRAIL_LENGTH;
    }
    
    // Counters for color value; start them at 0, we'll increment as we go along
    tmp_red = 0;
    tmp_green = 0;
    tmp_blue = 0;

    
    // Now actually draw the trail, from tail to head
    
    // Different drawing code if we're in DirectDraw mode; we separate it out here to pull a comparison and a little
    //  other overhead out (worth it because it's per-pixel and not per-spark beyond this point)
    if(dd_mode)
    {
      // Access the framebuffer directly, and draw based on colorspace
      //  The code here is very redundant, but we want to unroll the loops as much as possible
      //  for per-pixel drawing operations like this
      switch(colorspace)
      {
        // 32 bpp modes
        case B_RGB32:
        case B_RGBA32:
          tmp_buf_len = q_width * 4;

          for(j=0; j < sparks[i].trail_length; j++)
          {
            // Don't bother drawing this part of the spark trail if it's offscreen
            // Note: be careful that we include the width of the square being drawn when testing against the
            //  right and lower edges of the screen!
            if( ((sparks[i].trail_x[trail_pos]) < 0) || (sparks[i].trail_x[trail_pos] > (width-q_width))
             || ((sparks[i].trail_y[trail_pos]) < 0) || (sparks[i].trail_y[trail_pos] > (height-q_height)) )
            {
              // Skip drawing
            }
            else
            { 
              // Starting position in framebuffer memory
              tmp_pos = sparks[i].trail_ofs[trail_pos];
              
              // Draw the square
              // Setup pixel data buffer
              tmp_loop_pos = 0;
              for(x=0; x < q_width; x++)
              {
                tmp_buf[tmp_loop_pos++] = tmp_red;
                tmp_buf[tmp_loop_pos++] = tmp_green;
                tmp_buf[tmp_loop_pos++] = tmp_blue;
                tmp_loop_pos++;
              }
              
              // Copy data to framebuffer, for each row in square
              for(x=0; x < q_height; x++)
              {
                memcpy((void *)tmp_pos, tmp_buf, tmp_buf_len);
                tmp_pos += stride;
              }
            }
            // Increment the color values as we move up the tail towards the head  
            tmp_red += sparks[i].red;
            tmp_green += sparks[i].green;
            tmp_blue += sparks[i].blue;

            // Keep track of where we are in this spark's trail buffer
            trail_pos = ++trail_pos % MAX_TRAIL_LENGTH; 
          }
          break;

        // 24bpp mode
        case B_RGB24:
          tmp_buf_len = q_width * 3;

          for(j=0; j < sparks[i].trail_length; j++)
          {
            // Don't bother drawing this part of the spark trail if it's offscreen
            // Note: be careful that we include the width of the square being drawn when testing against the
            //  right and lower edges of the screen!
            if( ((sparks[i].trail_x[trail_pos]) < 0) || (sparks[i].trail_x[trail_pos] > (width-q_width))
             || ((sparks[i].trail_y[trail_pos]) < 0) || (sparks[i].trail_y[trail_pos] > (height-q_height)) )
            {
              // Skip drawing
            }
            else
            { 
              // Starting position in framebuffer memory
              tmp_pos = sparks[i].trail_ofs[trail_pos];
              
              // Draw the square
              // Setup pixel data buffer
              tmp_loop_pos = 0;
              for(x=0; x < q_width; x++)
              {
                tmp_buf[tmp_loop_pos++] = tmp_red;
                tmp_buf[tmp_loop_pos++] = tmp_green;
                tmp_buf[tmp_loop_pos++] = tmp_blue;
              }
              
              // Copy data to framebuffer, for each row in square
              for(x=0; x < q_height; x++)
              {
                memcpy((void *)tmp_pos, tmp_buf, tmp_buf_len);
                tmp_pos += stride;
              }
            }
            // Increment the color values as we move up the tail towards the head  
            tmp_red += sparks[i].red;
            tmp_green += sparks[i].green;
            tmp_blue += sparks[i].blue;

            // Keep track of where we are in this spark's trail buffer
            trail_pos = ++trail_pos % MAX_TRAIL_LENGTH; 
          }
          break;
          
        // 16 bpp mode
        case B_RGB16:
          tmp_buf_len = q_width * 2;

          for(j=0; j < sparks[i].trail_length; j++)
          {
            // Don't bother drawing this part of the spark trail if it's offscreen
            // Note: be careful that we include the width of the square being drawn when testing against the
            //  right and lower edges of the screen!
            if( ((sparks[i].trail_x[trail_pos]) < 0) || (sparks[i].trail_x[trail_pos] > (width-q_width))
             || ((sparks[i].trail_y[trail_pos]) < 0) || (sparks[i].trail_y[trail_pos] > (height-q_height)) )
            {
              // Skip drawing
            }
            else
            { 
              // Starting position in framebuffer memory
              tmp_pos = sparks[i].trail_ofs[trail_pos];
              
              // Draw the square
              // Pick colors for square
              c1 = blue_16_LUT[tmp_blue] | green_l_16_LUT[tmp_green];
              c2 = red_16_LUT[tmp_red] | green_h_16_LUT[tmp_green];
              
              // Setup pixel data buffer
              tmp_loop_pos = 0;
              for(x=0; x < q_width; x++)
              {
                tmp_buf[tmp_loop_pos++] = c1;
                tmp_buf[tmp_loop_pos++] = c2;
              }
              
              // Copy data to framebuffer, for each row in square
              for(x=0; x < q_height; x++)
              {
                memcpy((void *)tmp_pos, tmp_buf, tmp_buf_len);
                tmp_pos += stride;
              }
            }
            // Increment the color values as we move up the tail towards the head  
            tmp_red += sparks[i].red;
            tmp_green += sparks[i].green;
            tmp_blue += sparks[i].blue;

            // Keep track of where we are in this spark's trail buffer
            trail_pos = ++trail_pos % MAX_TRAIL_LENGTH; 
          }
          break;
          
        // 15bpp modes
        case B_RGB15:
        case B_RGBA15:
          tmp_buf_len = q_width * 2;

          for(j=0; j < sparks[i].trail_length; j++)
          {
            // Don't bother drawing this part of the spark trail if it's offscreen
            // Note: be careful that we include the width of the square being drawn when testing against the
            //  right and lower edges of the screen!
            if( ((sparks[i].trail_x[trail_pos]) < 0) || (sparks[i].trail_x[trail_pos] > (width-q_width))
             || ((sparks[i].trail_y[trail_pos]) < 0) || (sparks[i].trail_y[trail_pos] > (height-q_height)) )
            {
              // Skip drawing
            }
            else
            { 
              // Starting position in framebuffer memory
              tmp_pos = sparks[i].trail_ofs[trail_pos];
              
              // Draw the square
              // Pick colors for square
              // Blue here is the same as 16bpp mode
              c1 = blue_16_LUT[tmp_blue] | green_l_15_LUT[tmp_green];
              c2 = red_15_LUT[tmp_red] | green_h_15_LUT[tmp_green];
              
              // Setup pixel data buffer
              tmp_loop_pos = 0;
              for(x=0; x < q_width; x++)
              {
                tmp_buf[tmp_loop_pos++] = c1;
                tmp_buf[tmp_loop_pos++] = c2;
              }
              
              // Copy data to framebuffer, for each row in square
              for(x=0; x < q_height; x++)
              {
                memcpy((void *)tmp_pos, tmp_buf, tmp_buf_len);
                tmp_pos += stride;
              }
            }
            // Increment the color values as we move up the tail towards the head  
            tmp_red += sparks[i].red;
            tmp_green += sparks[i].green;
            tmp_blue += sparks[i].blue;

            // Keep track of where we are in this spark's trail buffer
            trail_pos = ++trail_pos % MAX_TRAIL_LENGTH; 
          }
          break;
            
        // Hmm, if we get here, something bad happened...
        default:
          break;
      }
    }

    // Not in DirectDraw mode, use the built-in BeOS drawing functions provided by the view for drawing
    else
    {
      // Start with the far end of the spark, then work forward to the head; this makes some
      //  situations when the spark's trail is bunched up on itself look better
      for(j=0; j < sparks[i].trail_length; j++)
      {
        // Don't bother drawing this part of the spark trail if it's offscreen
        // Note: be careful that we include the width of the square being drawn when testing against the
        //  right and lower edges of the screen!
        if( ((sparks[i].trail_x[trail_pos]) < 0) || (sparks[i].trail_x[trail_pos] > (width-q_width))
         || ((sparks[i].trail_y[trail_pos]) < 0) || (sparks[i].trail_y[trail_pos] > (height-q_height)) )
        {
          // Skip drawing
        }
        else
        { 
          // Set coordinates and size for drawing the square
          scratch_rect.Set(
            sparks[i].trail_x[trail_pos],
            sparks[i].trail_y[trail_pos],
            sparks[i].trail_x[trail_pos] + q_width,
            sparks[i].trail_y[trail_pos] + q_height
          );
          
          // Draw the square          
 
          // Make sure the tail is erased, not just blended 
          if( (blending) && (j == 0) )
          {
            view->SetDrawingMode(B_OP_COPY);
            view->SetHighColor(0, 0, 0);
            view->FillRect(scratch_rect);
            view->SetDrawingMode(B_OP_BLEND);
          }
          else
          {
            view->SetHighColor(tmp_red, tmp_green, tmp_blue);
            view->FillRect(scratch_rect);
          }

        }

        // Increment the color values as we move up the tail towards the head  
        tmp_red += sparks[i].red;
        tmp_green += sparks[i].green;
        tmp_blue += sparks[i].blue;

        // Keep track of where we are in this spark's trail buffer
        trail_pos = ++trail_pos % MAX_TRAIL_LENGTH; 
      }
    }
      

    // Check if spark is "dead" -- if it is, re-init it (offscreen)
    // (Spark is dead when it has timed out, and we've "burnt off" the entire length of the spark)
    if(sparks[i].timeout > 0)
    {
       sparks[i].timeout--;
    }
    else
    {
      if(sparks[i].trail_length > 0)
      {
        // We have to be sure to black-out the last square in the trail if we're going to shrink the spark's
        //  trail length for the "burn out" effect
        //  Separate code here for drawing in DirectDraw mode would be nice, but isn't really worth the extra
        //  mess in the source code; this gets called fairly rarely (only as a spark is burning out) in most
        //  cases anyway
        trail_pos = sparks[i].trail_counter - sparks[i].trail_length + 1;
        if(trail_pos < 0)
        {
          trail_pos += MAX_TRAIL_LENGTH;
        }
        trail_pos = trail_pos % MAX_TRAIL_LENGTH;
        scratch_rect.Set(
          sparks[i].trail_x[trail_pos],
          sparks[i].trail_y[trail_pos],
          sparks[i].trail_x[trail_pos] + q_width,
          sparks[i].trail_y[trail_pos] + q_height
        );
          
        // Make sure we're not using any blending
        if(blending)
        {
          view->SetDrawingMode(B_OP_COPY);
        }

        // Draw the square          
        view->SetHighColor(0, 0, 0);
        view->FillRect(scratch_rect);

        // Retore blending state
        if(blending)
        {
          view->SetDrawingMode(B_OP_BLEND);
        }

        // Now shrink the spark length
        sparks[i].trail_length--;
      }
      else
      {
        // Re-init spark...
        init_spark(i, width, height);
      }      
    }

  }

  // If in DirectDraw mode, unlock the screen since we're done drawing...
  if(dd_mode)
  {
    locker.Unlock();
      
    // And free up memeory for scratch buffer
    delete [] tmp_buf;
  }
  
  // If blending was on, turn it off
  if(blending)
  {
    view->SetDrawingMode(B_OP_COPY);
  }

}



// Initialize a spark - sets position, length, color values, etc.
void Sparticles::init_spark(int spark_num, float width, float height)
{
  int i;
  float a, b;
  int tmp_x, tmp_y, tmp_ofs;
  
  // Random scale of 0.000 to 1.000
  a = ((rand() % 1000) / 1000.0) * MAX_X_POS;
  b = ((rand() % 1000) / 1000.0) * MAX_Y_POS;


  // We want to create the sparks along the edges of the area - 10% margin outside the viewable area
  i = rand() % 4;
  
  switch(i)
  {
    // Left side
    case 0:
      sparks[spark_num].x = MIN_X_POS - (MAX_X_POS - MIN_X_POS) * 0.1;
      sparks[spark_num].y = b;
      break;

    // Right side
    case 1:
      sparks[spark_num].x = MAX_X_POS + (MAX_X_POS - MIN_X_POS) * 0.1;
      sparks[spark_num].y = b;
      break;

    // Top
    case 2:
      sparks[spark_num].x = a;
      sparks[spark_num].y = MIN_Y_POS - (MAX_Y_POS - MIN_Y_POS) * 0.1;
      break;

    // Bottom
    default:
      sparks[spark_num].x = a;
      sparks[spark_num].y = MAX_Y_POS + (MAX_Y_POS - MIN_Y_POS) * 0.1;
      break;
  }


  // Initial velocity is 0 (this will be modified almost immediately anyway)
  sparks[spark_num].velocity.x = 0.0;
  sparks[spark_num].velocity.y = 0.0;
          
  // Start at the beginning of the trail buffer, default length
  sparks[spark_num].trail_counter = 0;
  sparks[spark_num].trail_length = trail_length;

  // Set timeout -- amount of time before the spark "burns out"
  sparks[spark_num].timeout = (rand() % (MAX_TIMEOUT - MIN_TIMEOUT)) + MIN_TIMEOUT;

  // Set color values - these are actually the step-rates for changing the color as we draw from
  //  tail to head
  sparks[spark_num].red = (rand() % 256) / trail_length;
  sparks[spark_num].green = (rand() % 256) / trail_length;
  sparks[spark_num].blue = (rand() % 256) / trail_length;

  // Pre-load the trail buffer with the current position
  tmp_x = (int) ((sparks[spark_num].x/MAX_X_POS) * width);
  tmp_y = (int) ((sparks[spark_num].y/MAX_Y_POS) * height);
  tmp_ofs = (int) pixel_data + (tmp_x * 2) + (tmp_y * stride);
  for(i=0; i < MAX_TRAIL_LENGTH; i++)
  {
    // To make up some time, we do the actual on-screen coordinate calculation here
    sparks[spark_num].trail_x[i] = tmp_x;
    sparks[spark_num].trail_y[i] = tmp_y;
    sparks[spark_num].trail_ofs[i] = tmp_ofs;
  }
}


